const fs = require('fs-extra');
const path = require('path');
const url = require('url');
const dns = require('dns');

const puppeteer = require('puppeteer');
const moment = require('moment');
const archiver = require('archiver-promise');
const dateFormat = require('dateformat');
const whois = require('whois');
const whoisJSON = require('whois-json');
const sslCertificate = require('get-ssl-certificate');
const tldExtract = require('tld-extract');
const logger = require('log4js').getLogger('Scraper');
const EventEmitter = require('events');
const dig = require('./dig');
const jsonfile = require('jsonfile');
const TimeFormat = require('hh-mm-ss');

const { promisify } = require('util');

const whoisAsync = promisify(whois.lookup);

logger.level = 'debug';

const dnsServers = {
  'fr' : 'whois.afnic.fr:43',
  'cy' : 'whois.cynic.dns.cy:43'
};

const useDig = ['cy'];

class Scraper extends EventEmitter {

  constructor() {
    super(); // must call super for "this" to be defined.
    // this.dateTime = moment.calendarFormat("YYYYMMDD-HH-mm-ss");
    this.uriBase = process.env.SCRAPE_BASE_URI || 'https://register.fca.org.uk/ShPo_HomePage';
    this.nonrepudation = {};
    this.pathList = [];

    this.perf = {
      'started': 0,
      'finished': 0,
      'time': 0,
      'scraped': 0
    };
  }

  /**
   * 'Human' like click delay
   * @returns {number}
   */
  static notARobot() {
    return 90 + Math.floor(Math.random() * (30 - 1));
  }

  async _initBrowser() {
    this.browser = await puppeteer.launch({
      'executablePath': '/usr/bin/chromium-browser',
      'args': [
        // Use proxy so FCA wont block us
        `--proxy-server=${process.env.PROXY_URI}`,
        '--disable-dev-shm-usage',
        '--no-sandbox', '--headless', '--disable-gpu'
      ]
    });
  }

  async _makeScreenshot(page, destPath, waitFor = null) {
    if (waitFor) 
      await page.waitFor(waitFor);
    
    await page.setViewport({ 'width': 1200, 'height': 800 });
    await page.screenshot({ 'path': `artefacts/screenshots/${destPath}.png`, 'fullPage': true });
  }

  async _makeScreenshotV2(page, destPath, waitFor = null) {
    if (waitFor)
      await page.waitFor(waitFor);

    logger.debug('Snapshot', `${destPath}.png`);
    await page.setViewport({ 'width': 1200, 'height': 800 });
    await page.screenshot({ 'path': `${destPath}.png`, 'fullPage': true });
  }

  async _randomWait(page, minTime = 2, maxTime = 10) {
    const waitTime = Math.floor(Math.random() * (maxTime - minTime + 1) + minTime);
    logger.info(`Waiting ${waitTime} seconds...`);
    await page.waitFor(waitTime * 1000);
  }

  // use for artefacts saving only
  async _saveToFile(destPath, data) {
    return new Promise((resolve, reject) => {
      const fullPath = `${__dirname}/../artefacts/${destPath}`;
      fs.writeFile(fullPath, data, function(err) {
        if(err) 
          reject(err);
        else 
          resolve(`File saved to '${fullPath}'`);
      });
    });
  }

  async _dumpFile(destPath, data) {
    return new Promise((resolve, reject) => {
      fs.writeFile(destPath, data, function(err) {
        if(err)
          reject(err);
        else
          resolve(`File saved to '${destPath}'`);
      });
    });
  }

  async _createTimestampDirectory(destPath = null) {
    const now = new Date();

    const timestamp = dateFormat(now, 'yyyymmddHHMM');
    const fullPath = `${destPath}/${timestamp}`;

    logger.info('fullPath', fullPath);

    if (!fs.existsSync(fullPath))
      fs.ensureDirSync(fullPath);

    return fullPath;
  }

  /**
   *
   * @param destPath
   * @param filename
   * @returns {Promise<*>}
   * @private
   */
  async _createArchive(destPath = null, filename = null) {
    return new Promise((resolve, reject) => {
      if (!destPath || !filename) {
        const e = new Error('Missing paths');
        logger.error(e);
        reject(e);
      }
      var archive = archiver(filename, {
        'zlib': { 'level': 9 } // Sets the compression level.
      });

      archive.directory(`${destPath}/`);
      archive.finalize().then(function() {
        logger.debug('Archive finished');
        resolve();
      });
    });
  }

  /**
   *
   * @param urlStr
   * @returns {*}
   */
  explodeURL (urlStr = null) {
    if (!urlStr || urlStr === '')
      return (null);

    try {
      const workURL = url.parse(urlStr);

      return tldExtract.parse_host( workURL.host);
    }
    catch(e) {
      return e;
    }
  }

  /**
   *
   * @param destPath
   * @returns {Promise<void>}
   * @private
   */
  async _getWhoIsRaw(destPath = null) {
    const options = {};

    if (!destPath)
      reject(new Error('No destination path'));

    const explodedURL = this.explodeURL(destPath);

    if (dnsServers.hasOwnProperty(explodedURL.tld))
      options.server = dnsServers[explodedURL.tld];

    if (useDig.indexOf(explodedURL.tld) > -1)

      return await dig(`${explodedURL.sub}.${explodedURL.domain}`);

    else
      return await whoisAsync(`${explodedURL.domain}`, options);
  }

  /**
   *
   * @param destPath
   * @returns {Promise<*>}
   * @private
   */
  async _getWhoIsJSON(destPath = null) {
    const options = {};

    if (!destPath)
      reject(new Error('No destination path'));

    const explodedURL = this.explodeURL(destPath);

    if (dnsServers.hasOwnProperty(explodedURL.tld))
      options.server = dnsServers[explodedURL.tld];

    if (useDig.indexOf(explodedURL.tld) > -1)
      return {};

    else
      return await whoisJSON(`${explodedURL.domain}`, options);
  }

  /**
   *
   * @param destPath
   * @returns {Promise<*>}
   * @private
   */
  async _getWhoIsIPJSON(destPath = null) {
    return new Promise((resolve, reject) => {
      if (!destPath)
        reject(new Error('No destination path'));

      const workURL = url.parse(destPath);

      dns.lookup(workURL.host, (err, address, family) => {
        whoisJSON(address).catch((e) => {
          logger.error(e);
          reject(e);
        }).then((result) => {
          resolve(result);
        });
      });
    });
  }

  async _getWhoIsIPRaw(destPath = null) {
    return new Promise((resolve, reject) => {
      if (!destPath)
        reject(new Error('No destination path'));

      const workURL = url.parse(destPath);

      dns.lookup(workURL.host, (err, address, family) => {
        if (err)
          reject(err);

        whois.lookup(address, (err, data) => {
          if (err)
            reject(err);

          resolve(data);
        });
      });
    });
  }

  async _getSSLCert(destPath = null, prefix = false) {
    if (!destPath)
      return(new Error('No destination path'));

    const explodedURL = this.explodeURL(destPath);
    const searchFor = (prefix) ? `${explodedURL.sub}.${explodedURL.domain}` : `${explodedURL.domain}`;

    logger.debug('Cert for:', searchFor);
    
    return sslCertificate.get(searchFor, 5000);
  }
  
  async _populateNonRepudiation(destPath = null, options = {}) {
    this.nonrepudation.whois = {};
    this.nonrepudation.ipwhois = {};

    const sslWithPrefix = options.sslWithPrefix || false;

    logger.debug('Data for', destPath);

    await this._getWhoIsJSON(destPath).then((r) => {
      this.nonrepudation.whois.json = r;
    }).catch((err) => {
      logger.error(err);
    });

    await this._getWhoIsRaw(destPath).then((r) => {
      this.nonrepudation.whois.raw = r;
    }).catch((err) => {
      logger.error(err);
    });

    await this._getWhoIsIPRaw(destPath).then((r) => {
      this.nonrepudation.ipwhois.raw = r;
    }).catch((err) => {
      logger.error(err);
    });

    await this._getWhoIsIPJSON(destPath).then((r) => {
      this.nonrepudation.ipwhois.json = r;
    }).catch((err) => {
      logger.error(err);
    });

    await this._getSSLCert(destPath, sslWithPrefix).then((r) => {
      this.nonrepudation.sslcertificate = r;
    }).catch((err) => {
      logger.error(err);
    });

    if (this.nonrepudation.sslcertificate === null || typeof(this.nonrepudation.sslcertificate) === 'undefined') {
      logger.warn('Trying to retrieve SSL certificate with domain prefix.');
      await this._getSSLCert(destPath, true).then((r) => {
        this.nonrepudation.sslcertificate = r;
      }).catch((err) => {
        logger.error(err);
      });
    }
  }

  async _findAndClick(selector, text = null, url = null) {
    try {
      logger.debug('_findAndClick selector', selector);
      const clickableLinks = await this.page.$$(selector);
      const mouseDownDuration = Scraper.notARobot();

      let innerText;
      let href;
      if (clickableLinks.length > 0)
        for (const item of clickableLinks) {
          innerText = await this.page.evaluate(el => el.innerText, item);
          href = await this.page.evaluate(el => el.href, item);

          if( (text && innerText === text ) || (url && href === url )) {
            logger.debug('Matched item');
            await item.click({ 'delay':mouseDownDuration });
            await this._randomWait(this.page, 5, 10);
          }
          else
          if (!text && !url) {
            logger.debug('Just clicking element');
            await item.click({ 'delay':mouseDownDuration });
            await this._randomWait(this.page, 5, 10);
            // just click don't check
          }
        }
    }
    catch(err) {

      logger.error(err);
      process.exit(-99);
    }
    // selector `[id="${id}"] p a`
  }

  /**
   * Get Params from a url string
   */
  _getParamsFromUrl(url) {
    url = decodeURI(url);
    if (typeof url === 'string') {
      const params = url.split('?');

      const obj = {};
      if (params.length > 1) {
        const eachParamsArr = params[1].split('&');

        if (eachParamsArr && eachParamsArr.length)
          eachParamsArr.map(param => {
            const keyValuePair = param.split('=');
            const key = keyValuePair[0];
            const value = keyValuePair[1];
            obj[key] = value;
          });
      }

      return obj;
    }
  }

  _cleanUp(text) {
    if (!text) return;
    const regexNewLine = /\n/;
    const regexCollapseWS = /\s+/g;

    return text.replace(regexNewLine, '').replace(regexCollapseWS, ' ');
  }

  _start() {
    logger.debug('*** START ***');
    const now = new Date();
    this.perf.started = now.getTime();
  }

  _done() {
    logger.debug('*** DONE ***');
    const now = new Date();

    this.perf.finished = now.getTime();
    this.perf.duration = this.perf.finished - this.perf.started;

    this.perf.human = {};
    this.perf.human.duration = TimeFormat.fromMs(this.perf.duration, 'hh:mm:ss');

    jsonfile.writeFileSync(`${this.path}/perfdata.json`, this.perf);
  }
}

module.exports = Scraper;
