const Scraper = require('../helpers/scraper');

const path = require('path');
const jsonfile = require('jsonfile');
const logger = require('log4js').getLogger('CY');

logger.level = 'debug';

// load env variables from file

require('dotenv').config({
  'path': `${__dirname }/../.env`
});

class CYScrape extends Scraper {

  /**
   *
   * @param selector
   * @returns {Promise<void>}
   */
  async grabLink(selector) {
    const clickableLinks = await this.page.$$(selector);

    const tsDirectory = await this._createTimestampDirectory(this.path);
    await this.page._client.send('Page.setDownloadBehavior', { 'behavior': 'allow', 'downloadPath': tsDirectory });

    if (clickableLinks.length > 0)
      for (const item of clickableLinks) {
        const href = await this.page.evaluate(el => el.href, item);
        await this._randomWait(this.page, 3, 5);
        await this.page.goto(href, { 'waitUntil': 'networkidle2' }).catch((err) => {
          // log this error but Puppeteer isn't supposed to support this sort of download....
          logger.error(err);
        });
      }
  }

  /**
   *
   * @param id
   * @returns {Promise<void>}
   */
  async downloadEmoney(id) {
    const selector = ['#generic_article > div > div.row > div > div > ul > li:nth-child(1) > a', '#generic_article > div > div.row > div > div > ul > li:nth-child(2) > b > b > a'];

    this.grabLink(selector[id]);
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async downloadExcel() {
    const selector = '#workshops > div > div.workshop-article-container > div > div > div > h3 > a';

    this.grabLink(selector);
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async handlePaymentInstitutions() {
    await this._randomWait(this.page, 3, 5);

    const filename = 'licensing-and-supervision-of-payment-institutions';

    const tsDirectory = await this._createTimestampDirectory(this.path);

    this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_main`, null);
    await this._randomWait(this.page, 3, 5);

    await this.downloadExcel();
    await this._randomWait(this.page, 3, 5);

    await this.page.goto(this.eMoneyUrl, { 'waitUntil': 'networkidle2' });
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async handleElectronicMoneyInstitutions() {
    await this._randomWait(this.page, 3, 5);

    const filename = 'licensing-and-supervision-of-electronic-money-institutions';

    const tsDirectory = await this._createTimestampDirectory(this.path);

    this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_main`, null);
    await this._randomWait(this.page, 3, 5);

    await this.downloadEmoney(0);
    await this._randomWait(this.page, 3, 5);

    await this.downloadEmoney(1);
    await this._randomWait(this.page, 3, 5);
    this.emit('done');
  }

  /**
   *
   * @param filePath
   * @returns {Promise<void>}
   */
  async savePDF(filePath) {
    logger.info('Saving the pdf:', filePath);

    await this._randomWait(this.page, 5, 7);
    await this.page.pdf({ 'path': filePath, 'format': 'A4' });
    this.emit('done');
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processNewPage() {
    // give the page a few seconds to settle
    const checkPDF = /(.pdf)/g;

    await this._randomWait(this.page, 3, 5);

    const currentPage = await this.page.evaluate(() => document);

    let currentPath = currentPage.location.pathname;
    let pdfFile;

    if (checkPDF.test(currentPath)) {
      const splitPath = currentPath.split('/');

      pdfFile = splitPath.pop();
      currentPath = splitPath.join('/');
    }

    switch (currentPath) {

      case '/en/licensing-supervision/payment-institutions/licensing-and-supervision-of-payment-institutions':
        await this.handlePaymentInstitutions();
        break;
      case '/en/licensing-supervision/electronic-money-institutions/licensing-and-supervision-of-electronic-money-institutions':
        await this.handleElectronicMoneyInstitutions();
        break;
      case '/images/media/redirectfile/Electronic%20Money%20Institutions':
        logger.warn('We should only arrive here when in Non-headless mode');
        await this.savePDF(pdfFile);
        break;
      default:
        logger.debug('processNewPage Fell through');
        logger.debug('currentPath', currentPath);

        break;

    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async start() {
    try {
      this.startPage = 'https://www.centralbank.cy/en/licensing-supervision/payment-institutions/licensing-and-supervision-of-payment-institutions';
      this.eMoneyUrl = 'https://www.centralbank.cy/en/licensing-supervision/electronic-money-institutions/licensing-and-supervision-of-electronic-money-institutions';

      this.path = path.resolve(`${__dirname }/../artefacts/CY/CBOC`);

      await this._populateNonRepudiation(this.startPage);

      const tsDirectory = await this._createTimestampDirectory(this.path);
      this.pathList.push(tsDirectory);

      await jsonfile.writeFileSync(`${tsDirectory}/nonrepudiation.json`, this.nonrepudation);
      await this._initBrowser();

      this.page = await this.browser.newPage();

      this.page.on('domcontentloaded', () => {
        this.processNewPage();
      });

      await this.page.tracing.start({ 'path': `${this.path}/trace.json`, 'screenshots': true });

      await this.page.setViewport({ 'width': 1200, 'height': 800 });
      await this.page.goto(this.startPage);

      await this._randomWait(this.page, 3, 5);
    }
    catch (e) {
      throw new Error(e);
    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async run() {
    logger.info('Scraping Cyprus...');

    this.on('done', () => {
      logger.warn('Proper done');
      this.page.tracing.stop().catch((err) => {
        logger.error(err);
      }).then(() => {
        // logger.debug(this);
        this.browser.close();
      });
    });

    await this.start();
  }
}

module.exports = CYScrape;
