const Scraper = require('../helpers/scraper');
const cheerio = require('cheerio');
const path = require('path');
const jsonfile = require('jsonfile');
const logger = require('log4js').getLogger('DE');

logger.level = 'debug';

// load env variables from file

require('dotenv').config({
  'path': `${__dirname }/../.env`
});

class DEScrape extends Scraper {

  /**
   *
   * @returns {Promise<void>}
   */
  async buildSubIndex() {
    logger.info('Building sub-index...');

    const currentPage = await this.page.evaluate(() => document);

    const search = currentPage.location.search;
    const params = this._getParamsFromUrl(search);

    const currentPageID = params.nameZahlungsinstitut || '';

    const tsDirectory = await this._createTimestampDirectory(this.path);

    await this._makeScreenshotV2(this.page, `${tsDirectory}/menu_${currentPageID}`, null);

    await this._randomWait(this.page, 3, 5);

    const links = await this.page.$$('#zahlinst a');

    for (const item of links) {
      const id = await this.page.evaluate(el => el.innerText, item);
      let href = await this.page.evaluate(el => el.href, item);

      href = href.concat('&locale=en_GB');

      if (id !== 'Found payment institutions:') 
        this.paymentServices.links.push({ id, href });
    }

    this.index.step++;
    
    if (this.index.step < this.index.items) 
      this.emit('nextsubindex');
    else {
      this.subIndex.done = true;
      this.paymentServices.items = this.paymentServices.links.length;
      this.emit('subindexdone');
    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async buildIndex() {
    logger.info('Building the index...');

    await this._randomWait(this.page, 3, 5);

    const links = await this.page.$$('#suchform > div > div:nth-child(2) > div.navigationGruppeBuchstaben a');

    for (const item of links) {
      const id = await this.page.evaluate(el => el.innerText, item);
      let href = await this.page.evaluate(el => el.href, item);

      href = href.concat('&locale=en_GB');

      this.index.links.push({ id, href });
    }

    this.index.done = true;
    this.index.items = this.index.links.length;

    this.emit('indexdone');
  }

  /**
   *
   * @param html
   * @returns {Promise<{description: T[] | jQuery, permissions: {original: Array, translated: Array}}>}
   */
  async extractPaymentEntity(html) {
    const permissions = { 'original':[], 'translated':[] };

    const newLine = /\n/g;
    const $ = cheerio.load(html);

    const description = $('#content > p').text().split(newLine).filter(line => line.length > 0);

    const rows = $('#erlaubnis > tbody tr');
    
    rows.each((index, item) => {
      const cells = $(item).find('td');

      const service = $(cells.get(0)).text();
      const startAuth = $(cells.get(1)).text();
      const endAuth = $(cells.get(2)).text();

      const phrasing = service.split(' (§');

      if (!this.dictionary.get(phrasing[0])) 
        this.dictionary.set(phrasing[0], '');
      else 
        phrasing[0] = this.dictionary.get(phrasing[0]);

      permissions.translated.push({
        'service': phrasing.join(' (§'),
        startAuth,
        endAuth
      });

      permissions.original.push({
        service,
        startAuth,
        endAuth
      });
    });

    return { description, permissions };
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processEntity() {
    const noWhiteSpace = /\W/g;
    if (!this.subIndex.done) {
      // We should not be here quite yet, so add this to subindex;
      const currentPage = await this.page.evaluate(() => document);

      const location = currentPage.location;
      const id = location.search;
      let href = location.href;
      href = href.concat('&locale=en_GB');

      this.paymentServices.links.push({ id, href });

      this.index.step++;

      if (this.index.step < this.index.items)
        this.emit('nextsubindex');
      else {
        logger.info('Sub indexing done...');
        this.subIndex.done = true;
        this.paymentServices.items = this.paymentServices.links.length;
        this.emit('subindexdone');
      }
    }
    else {
      const id = this.paymentServices.links[this.paymentServices.step].id;
      logger.info('Process entity:', id);
      await this._randomWait(this.page, 3, 5);

      const body = await this.page.evaluate(() => document.documentElement.outerHTML);

      const details = await this.extractPaymentEntity(body);

      const filename = id.indexOf('?id=') === 0 ? `${details.description[0].replace(noWhiteSpace, '_')}` : `${id.replace(noWhiteSpace, '_')}`;

      const tsDirectory = await this._createTimestampDirectory(this.path);

      await this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_main`, null);

      jsonfile.writeFileSync(`${tsDirectory}/${filename}.json`, details);

      this.paymentServices.step++;

      if (this.paymentServices.step < this.paymentServices.items) 
        await this.page.goto(this.paymentServices.links[this.paymentServices.step].href);
      else 
        this.emit('processdone');
    }
  }

  /**
   *
   * @param selector
   * @returns {Promise<void>}
   */
  async grabLink(selector) {
    const clickableLinks = await this.page.$$(selector);

    const tsDirectory = await this._createTimestampDirectory(this.path);
    await this.page._client.send('Page.setDownloadBehavior', { 'behavior': 'allow', 'downloadPath': tsDirectory });

    if (clickableLinks.length > 0)
      for (const item of clickableLinks) {
        const href = await this.page.evaluate(el => el.href, item);
        await this._randomWait(this.page, 3, 5);
        await this.page.goto(href, { 'waitUntil': 'networkidle2' }).catch((err) => {
          // log this error but Puppeteer isn't supposed to support this sort of download....
          logger.warn(err);
        });
      }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processEMoney() {
    logger.info('Process EMoney:');
    await this._randomWait(this.page, 3, 5);

    const filename = 'e-money_Institutions';

    const tsDirectory = await this._createTimestampDirectory(this.path);

    await this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_main`, null);

    await this._findAndClick('#content > div.sectionRelated.toggleEntry.gsb-toggle > div > h3:nth-child(5)');

    await this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_expanded`, null);

    await this.grabLink('#content > div.sectionRelated.toggleEntry.gsb-toggle > div > ul:nth-child(6) > li > a');

    await this._randomWait(this.page, 3, 5);

    this.emit('done');
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processNewPage() {
    // give the page a few seconds to settle
    await this._randomWait(this.page, 3, 5);

    const currentPage = await this.page.evaluate(() => document);

    const search = currentPage.location.search;

    switch (currentPage.location.pathname) {

      case '/database/ZahlInstInfo/':
        await this.buildIndex();
        break;

      case '/database/ZahlInstInfo/suche.do':
        await this.buildSubIndex();
        break;
      case '/database/ZahlInstInfo/zahlinst.do':
        await this.processEntity();
        break;
      case '/DE/PublikationenDaten/Datenbanken/EGeldInstitute/e-geld-institute_node.html':
        await this.processEMoney();
        break;
      default:
        logger.debug('processNewPage Fell through');
        logger.debug('currentPage.location', currentPage.location);
        logger.debug('Search', search);
        break;

    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async start() {
    try {
      const dictionary = jsonfile.readFileSync('helpers/dictionary/de.json');

      this.dictionary = new Map(dictionary);

      this.index = {
        'items': 0,
        'links': [],
        'step': 0,
        'started': false,
        'done' : false
      };

      this.subIndex = {
        'items': 0,
        'links': [],
        'step': 0,
        'started': false,
        'done' : false
      };

      this.paymentServices = {
        'items': 0,
        'links': [],
        'step': 0,
        'visited': false,
        'done' : false
      };

      this.startPage = 'https://portal.mvp.bafin.de/database/ZahlInstInfo/?locale=en_GB';
      this.emoneyUrl = 'https://www.bafin.de/DE/PublikationenDaten/Datenbanken/EGeldInstitute/e-geld-institute_node.html';

      this.path = path.resolve(`${__dirname }/../artefacts/DE/BAFIN`);

      await this._populateNonRepudiation(this.startPage);

      const tsDirectory = await this._createTimestampDirectory(this.path);
      this.pathList.push(tsDirectory);

      await jsonfile.writeFileSync(`${tsDirectory}/nonrepudiation.json`, this.nonrepudation);

      await this._initBrowser();
      this.page = await this.browser.newPage();

      this.page.on('domcontentloaded', () => {
        this.processNewPage();
      });

      this.on('processdone', async function() {
        logger.debug('Payment Entities done', this.paymentServices.items);

        jsonfile.writeFileSync('helpers/dictionary/de.json', [...this.dictionary]);
        await this._randomWait(this.page, 5, 10);
        await this.page.goto(this.emoneyUrl);
      });

      this.on('subindexdone', async function() {
        logger.debug('Sub Index done', this.paymentServices.items);
        logger.debug(this.paymentServices.links[this.paymentServices.step].href);
        await this.page.goto(this.paymentServices.links[this.paymentServices.step].href);
        logger.debug('After..');
      });

      this.on('indexdone', async function() {
        logger.debug('Index done', this.index.items);
        logger.debug(this.index.links[this.index.step].href);
        await this.page.goto(this.index.links[this.index.step].href);
        logger.debug('After..');
      });

      this.on('nextsubindex', async function() {
        logger.debug(this.index.links[this.index.step].href);
        await this.page.goto(this.index.links[this.index.step].href);
      });

      await this.page.tracing.start({ 'path': `${this.path}/trace.json`, 'screenshots':true }).catch((err) => {
        logger.error(err);
      });

      await this.page.setViewport({ 'width': 1200, 'height': 800 });
      await this.page.goto(this.startPage);

      await this._randomWait(this.page, 3, 5);
    }
    catch(e) {
      throw new Error(e);
    }
  }

  async run() {
    await this.start();

    this.on('done', () => {
      logger.warn('Proper done');
      this.page.tracing.stop().catch((err) => {
        logger.error(err);
      }).then(() => {
        // logger.debug(this);
        this.browser.close();
      });
    });

    await this.page.tracing.stop();
  }

}

module.exports = DEScrape;
