const Scraper = require('../helpers/scraper');
const cheerio = require('cheerio');
const path = require('path');
const jsonfile = require('jsonfile');
const logger = require('log4js').getLogger('FR');
const url = require('url');
const removeAccents = require('remove-accents-diacritics');

const EventEmitter = require('events');

logger.level = 'debug';

// load env variables from file

require('dotenv').config({
  'path': `${__dirname }/../.env`
});

class FRScrape extends Scraper {

  constructor() {
    super(); // must call super for "this" to be defined.

    this.on('done', () => {
      logger.warn('Scraping complete:');

      this._done();

      this.browser.close().catch((err) => {
        logger.error(err);
      });
    });
  }

  /**
   *
   * @param path
   * @returns {Promise<void>}
   */
  async gotoPage(path = null) {
    const newurl = `${this.parsedUrl.protocol}//${this.parsedUrl.hostname}${path.link}`;
    await this._randomWait(this.page, 3, 5);

    logger.debug('newurl:', newurl);
    await this.page.goto(newurl);
  }

  /**
   *
   * @param rows
   * @returns {Array}
   */
  extractDataFromTable(rows) {
    const unchecked = /(unchecked)/;
    const output = [];
    const crossBorder = [];

    let currentActivityID ;
    rows.each((i, elm) => {
      const children = cheerio(elm).children();
      let newItem;

      if (children.eq(1).text().trim() !== '')
        currentActivityID = children.eq(1).text().trim();

      if (children.eq(0).html().match(unchecked) === null)
        if (children.length === 2) {
          crossBorder.push(this._cleanUp(currentActivityID.trim()));
        }
        else
        if (children.length === 3) {
          newItem = [currentActivityID, this._cleanUp(children.eq(2).text().trim())];
          output.push(newItem);
        }
        else {
          newItem = [`${currentActivityID}${children.eq(2).text().replace(')', '').trim()}`, this._cleanUp(children.eq(3).text().trim())];
          output.push(newItem);
        }
    });

    return { output, crossBorder };
  }

  extractDataFromInvestmentServicesTable(rows) {
    const unchecked = /(unchecked)/;
    const cellChecker = /L([0-9]{1,3})C([0-9]{1,3})/;
    const output = [];
    const authorised = [];
    const financialInstruments = [];

    rows.each((i, elm) => {
      const finInst = [];
      const children = cheerio(elm).children();

      if (children.length > 2) {
        if (children.length === 11) 
          children.each((step, fiElm) => {
            financialInstruments.push(this._cleanUp(cheerio(fiElm).text()));
          });

        if (children.length > 11) {
          let offset = (children.length - 1) - financialInstruments.length;
          const fiOffset = (offset === 0) ? 1 : 2;

          const rowName = children.eq(offset).text();
          offset++;
          while(offset < financialInstruments.length) {
            if (children.eq(offset).html().match(unchecked) === null)
              finInst.push(financialInstruments[offset - fiOffset]);
            
            offset++;
          }
          if (finInst.length > 0) 
            output.push([rowName, finInst]);
        }
      }
      else if (children.length === 2) 

        if (children.eq(0).html().match(unchecked) === null) {
          authorised.push(this._cleanUp(children.eq(1).text()));
        }
    });

    return { 'investmentServices':output, authorised };
  }

  /**
   *
   * @param tables
   * @returns {Promise<Array>}
   */
  async extractEuroData(tables) {
    const dataBlock = [];
    const findToColon = /^.*?(?=(:))/;
    const trimToColon = /^.*?(?=(:)).\s/;

    const divs = tables.find('div.zone_succ');

    divs.each((i, elm) => {
      const p = cheerio(elm).find('p').eq(0).text();

      const title = p.match(findToColon)[0];
      const country = p.split(trimToColon)[2];

      const obj = {};
      obj[title] = country;

      const rows = cheerio(elm).find('table tr');

      const data = this.extractDataFromTable(rows);

      obj.paymentServices = data.output;
      obj.crossBorder = data.crossBorder;

      dataBlock.push(obj);
    });

    return dataBlock;
  }

  async extractLinks($table, creditInstFilter = false) {
    const wantedCIStatuses = ['legal entity/ company'];
    const links = [];
    if ($table.length > 1)
    // The table contains more than just the heading row
      for (let count = 1;count < $table.length;count++) {
        const $row = cheerio($table.get(count)).find('td');

        const $item = $row.children().eq(2);

        const link = $item.attr('href');
        const title = this._cleanUp($item.text());

        if (!creditInstFilter)
          // Default mode
          links.push({ link, title });
        else 
        if ($row.children().length === 6) {
          const status = this._cleanUp($row.children().eq(5).text().toLowerCase());

          if(wantedCIStatuses.indexOf(status) !== -1) 
            links.push({ link, title });
        }
      }

    return links;
  }

  /**
   *
   * @param $
   * @returns {Promise<Array>}
   */
  async extractDetails($) {
    const findToColon = /^.*?(?=(:))/;
    const trimToColon = /^.*?(?=(:)).\s/;
    const details = [];

    $('div#zone_description ul.nopuce li').each((i, elm) => {
      if ($(elm).children().length > 0) {
        const matched = $(elm).text().match(findToColon);

        if (matched !== null) {
          const field = $(elm).text().match(findToColon)[0].trim();
          const data = this._cleanUp( $(elm).text().split(trimToColon)[2]);

          details.push([field, data]);
        }
      }
    });
    
    return details;
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processAFPage() {
    const noWhiteSpace = /\W/g;
    const trimToColon = /^.*?(?=(:)).\s/;

    const body = await this.page.evaluate(() => document.documentElement.outerHTML);
    const $ = cheerio.load(body);
    const modeFilename = ['', 'eMoney_', 'credit_'];

    const pageData = { 'description':[], 'frActivities':null, 'EUActivities':[] };

    pageData.entity = removeAccents.remove($('p.sttr').eq(0).text().replace(trimToColon, '').trim());

    const filename = `${modeFilename[this.mode]}${pageData.entity.replace(noWhiteSpace, '_')}`;

    const tsDirectory = await this._createTimestampDirectory(this.path);

    this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_main`, null);

    pageData.description = await this.extractDetails($);

    await this._findAndClick('div.main.main_evol > table > tbody > tr > td:nth-child(3) a');

    // Process France / French details

    this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_france`, null);

    const frenchTbl = $('#zone_en_france > table tr');

    if (this.mode < 2) 
      pageData.frActivities = await this.extractDataFromTable(frenchTbl).output;
    else 
      pageData.creditInstituteActivities = await this.extractDataFromInvestmentServicesTable(frenchTbl);

    if (this.mode < 2) {
      await this._findAndClick('div.main.main_evol > table > tbody > tr > td:nth-child(5) a');

      // Process EU Details

      this._makeScreenshotV2(this.page, `${tsDirectory}/${filename}_europe`, null);

      const euroTbls = $('#zone_en_europe');

      pageData.EUActivities = await this.extractEuroData(euroTbls);
    }

    jsonfile.writeFileSync(`${tsDirectory}/${filename}.json`, pageData);

    if (this.mode === 0 )
      this.paymentServices.step++;
    else if( this.mode === 1)
      this.emoneyServices.step++;
    else if( this.mode === 2)
      this.creditServices.step++;

    this.perf.scraped++;
    await this._randomWait(this.page, 5, 7);

    if (this.mode === 0)
      if (this.paymentServices.step < this.paymentServices.items)
        await this.gotoPage(this.paymentServices.links[this.paymentServices.step]);
      else {
        logger.debug('Payment services complete.');
        this.paymentServices.done = true;
        this.mode++;

        await this.page.goto(this.eMoneyUrl);
      }

    else if (this.mode === 1)
      if (this.emoneyServices.step < this.emoneyServices.items)
        await this.gotoPage(this.emoneyServices.links[this.emoneyServices.step]);
      else {
        logger.debug('EMoney services complete.');
        this.emoneyServices.done = true;
        this.mode++;

        await this.page.goto(this.creditUrl);
      }

    else if (this.mode === 2)
      if (this.creditServices.step < this.creditServices.items)
        await this.gotoPage(this.creditServices.links[this.creditServices.step]);
      else {
        logger.debug('Credit services complete.');
        this.creditServices.done = true;
        this.mode++;

        this.emit('done');
      }
  }

  /**
   *
   * @param $
   * @param store
   * @returns {Promise<void>}
   */
  async searchResultsProcessor($, store) {
    const $table = $('table.table tr');

    if ($table.length > 1)
    // The table contains more than just the heading row

      store.links = store.links.concat(await this.extractLinks($table, (this.mode === 2)));

    // check that the next button is active

    const nextExists = $('body > div > div.main.main_evol > ul > li:last-child > a');

    if (nextExists.length === 1 )
      await this._findAndClick('body > div > div.main.main_evol > ul > li:last-child > a', 'Next page >');
    else {
      // Done gathering search results
      logger.info('Completed gathering search results..');
      store.searchDone = true;
      store.items = store.links.length;

      jsonfile.writeFileSync(`${this.path}/${['pi', 'eu', 'ci'][this.mode]}.json`, store);

      this.gotoPage(store.links[store.step]);
    }
  }

  /**
   * Handle the search result page and uilt the list of links
   * @returns {Promise<void>}
   */
  async handleSearchResults() {
    const body = await this.page.evaluate(() => document.documentElement.outerHTML);
    const $ = cheerio.load(body);

    if (this.mode === 0 && !this.paymentServices.searchDone)
      await this.searchResultsProcessor($, this.paymentServices);

    if (this.mode === 1 && !this.emoneyServices.searchDone)
      await this.searchResultsProcessor($, this.emoneyServices);

    if (this.mode === 2 && !this.creditServices.searchDone)
      await this.searchResultsProcessor($, this.creditServices);
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async processNewPage(dump = false) {
    // give the page a few seconds to settle
    await this._randomWait(this.page, 3, 5);

    const currentPage = await this.page.evaluate(() => document);

    logger.debug(await this.page.evaluate(() => document.documentElement.outerHTML));

    if (dump) 
      await this._dumpFile(`${this.path}/${now.getTime()}.html`, await this.page.evaluate(() => document.documentElement.outerHTML));

    const search = currentPage.location.search;
    const params = this._getParamsFromUrl(search);
    const pageID = params.page || '';
    switch (pageID) {

      case 'results':
        await this.handleSearchResults( params);
        break;
      case 'af':
        await this.processAFPage();
        break;

      default:
        logger.debug('processNewPage Fell through');
        logger.debug('Path', currentPage.location.pathname);
        logger.debug('Search', search);
        logger.debug('Params', this._getParamsFromUrl(search));
        break;

    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async start() {
    super._start();
    try {
      this.mode = 0;

      this.paymentServices = {
        'items': 0,
        'links': [],
        'step': 0,
        'visited': false,
        'done' : false,
        'searchDone' : false
      };

      this.emoneyServices = {
        'items': 0,
        'links': [],
        'step': 0,
        'visited': false,
        'done' : false,
        'searchDone' : false
      };

      this.creditServices = {
        'items': 0,
        'links': [],
        'step': 0,
        'visited': false,
        'done' : false,
        'searchDone' : false
      };

      this.startPage = 'https://www.regafi.fr/spip.php?page=results&type=advanced&denomination=&siren=&cib=&bic=&nom=&siren_agent=&num=&cat=21-TBR07&retrait=0&lang=en&id_secteur=3';
      this.eMoneyUrl = 'https://www.regafi.fr/spip.php?page=results&type=advanced&id_secteur=3&lang=en&denomination=&siren=&cib=&bic=&nom=&siren_agent=&num=&cat=22-TBR07&retrait=0';
      this.creditUrl = 'https://www.regafi.fr/spip.php?page=results&type=advanced&id_secteur=3&lang=en&denomination=&siren=&cib=&bic=&nom=&siren_agent=&num=&cat=01-TBR07&retrait=0';

      this.parsedUrl = url.parse(this.creditUrl);
      this.path = path.resolve(`${__dirname }/../artefacts/FR/REGAFI`);

      await this._populateNonRepudiation(this.startPage, { 'sslWithPrefix':true });

      const tsDirectory = await this._createTimestampDirectory(this.path);
      this.pathList.push(tsDirectory);

      await jsonfile.writeFileSync(`${tsDirectory}/nonrepudiation.json`, this.nonrepudation);
      await this._initBrowser();
      this.page = await this.browser.newPage();

      this.page.on('domcontentloaded', () => {
        this.processNewPage();
      });

      await this.page.tracing.start({ 'path': `${this.path}/trace.json`, 'screenshots':true });

      await this.page.setViewport({ 'width': 1200, 'height': 800 });
      await this.page.goto(this.startPage);

      await this._randomWait(this.page, 3, 5);
    }
    catch(e) {
      throw new Error(e);
    }
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async run() {
    logger.info('Scraping France...');

    await this.start();
  }

}

module.exports = FRScrape;
