/**
 *
 * User: Martin Donnelly
 * Date: 2018-09-13
 * Time: 12:23
 *
 */

const Scraper = require('../helpers/scraper');
const cheerio = require('cheerio');
const path = require('path');
const jsonfile = require('jsonfile');

// load env variables from file

require('dotenv').config({
  'path': `${__dirname }/../.env`
});

class IEScrape extends Scraper {

  constructor() {
    super();
  }

  /**
   *
   * @returns {Promise<void>}
   */
  async start() {
    try{
      this.startPage = 'http://registers.centralbank.ie/Home.aspx';
      const mouseDownDuration = IEScrape.notARobot();
      this.path = path.resolve(`${__dirname }/../artefacts/IE/CBI`);

      await this._populateNonRepudiation(this.startPage);

      const tsDirectory = await this._createTimestampDirectory(this.path);
      this.pathList.push(tsDirectory);

      await jsonfile.writeFileSync(`${tsDirectory}/nonrepudiation.json`, this.nonrepudation);

      await this._initBrowser();
      this.page = await this.browser.newPage();

      await this.page.tracing.start({ 'path': `${this.path}/trace.json`, 'screenshots':true });

      await this.page.setViewport({ 'width': 1200, 'height': 800 });
      await this.page.goto(this.startPage);

      await this._randomWait(this.page, 3, 5);
      await this.page.waitForSelector('#ctl00_cphRegistersMasterPage_lblViewList');
      await this.page.click('#ctl00_cphRegistersMasterPage_lblViewList > a', { 'delay':mouseDownDuration });
    }
    catch(e) {
      throw new Error(e);
    }
  }

  /**
   * locate the download section associated with the searchText
   * @param downloadsection
   * @param searchText
   * @returns {Promise<*>}
   */
  async findDownloadSection(downloadsection, searchText) {
    try{
      await this.page.waitFor(downloadsection);

      const body = await this.page.evaluate(() => document.documentElement.outerHTML);
      const $ = cheerio.load(body);

      let wantedId;

      $(`${downloadsection} span`).each((i, el) => {
        if ($(el).text() === searchText)
          wantedId = $(el).attr('id');

        return wantedId;
      });

      return wantedId;
    }
    
    catch(e) {
      throw new Error(e);
    }
  }

  /**
   * Expand the relevant section
   * @param elmId
   * @returns {Promise<void>}
   */
  async expandArea(elmId) {
    await this.page.click(`span#${elmId}`);
  }

  /**
   * Find the Download Links via section ID
   * @param elmId
   * @returns {Promise<void>}
   */
  async findDownloadsLinksID(elmId) {
    return await this.page.$eval(`span#${elmId}`, e => e.parentElement.nextElementSibling.getAttribute('id'));
  }

  /**
   * Process the download links and grab the pdf files
   * @param id
   * @returns {Promise<void>}
   */
  async processDownloadLinks(id) {
    // Each link is duplicated in a P and an Image. We just use the one in the P tag.

    const clickableLinks = await this.page.$$(`[id="${id}"] p a`);
    const mouseDownDuration = IEScrape.notARobot();

    let tsDirectory;
    for (const item of clickableLinks) {
      tsDirectory = await this._createTimestampDirectory(this.path);

      this.pathList.push(tsDirectory);

      await this.page._client.send('Page.setDownloadBehavior', { 'behavior': 'allow', 'downloadPath': tsDirectory });

      await item.click({ 'delay':mouseDownDuration });
      await this._randomWait(this.page, 5, 10);
    }
  }

  async grabSection(dlSectionElm, sectionTitle) {
    const section = await this.findDownloadSection(dlSectionElm, sectionTitle);

    await this.expandArea(section);

    const tsDirectory = await this._createTimestampDirectory(this.path);
    this.pathList.push(tsDirectory);
    this._makeScreenshotV2(this.page, `${tsDirectory}/${sectionTitle}`, null);

    const sectionID = await this.findDownloadsLinksID(section);

    await this.processDownloadLinks(sectionID);

    await this._randomWait(this.page, 5, 10);
  }

  /**
   * Grab the Pdf's and screenshots
   * @returns {Promise<void>}
   */
  async run() {
    await this.start();

    await this._randomWait(this.page, 5, 10);
    const tsDirectory = await this._createTimestampDirectory(this.path);

    this.pathList.push(tsDirectory);
    this._makeScreenshotV2(this.page, `${tsDirectory}/Central Bank of Ireland Registers`, null);

    const sections = ['Registers of Payment Services Firms', 'Registers of E-Money Firms', 'Register of Credit Institutions'];

    for (const section of sections)
      await this.grabSection('#ctl00_cphRegistersMasterPage_downloadsSection', section);

    await this._createArchive('artefacts', './ie.zip');

    await this.page.tracing.stop();

    await this.browser.close();

    return this.pathList;
  }

}

module.exports = IEScrape;
